******************************************************
*  Program...........: L7ParserSamples.prg
*  Author............: Lauren Clarke
*  Project...........: L7
*  Created...........: 05/30/2002
*  Copyright.........: Cycla Corp with use granted to Cornerstone Systems Inc.
*  Description.......: Various text to html parsers for general <textarea to memo use
*  Unit Test.........: L7Parsers_ut.prg
*  Dependencies......: VFP7, RELATION, L7Utils, VBScript.RegExp (or equiv)
*  Assumptions.......: SET EXACT OFF    
*  Last Change.......: 12/30/2002
*  Change Log........: See end of file
******************************************************
* COMMENTS
* This file contains the core L7Parser base-classes and 
* a selection of  subclasses for demonstration purposes
*
* USAGE
* See unit test for usage
*	

#DEFINE L7_PARSER_PLACEHOLDER_DELIM CHR(222)
#DEFINE L7_PARSER_DEBUG_OBJECTS .f.
#DEFINE L7_PARSER_DEBUG .f. 
#DEFINE L7_PARSER_PERFLOG .f.
#DEFINE L7_REGEXP_CLASS "vbScript.regexp"



*** ========================================================== ***
DEFINE CLASS L7TextDecorationParser as L7MultiParser
* provides basic plain-text to html wysiwyg pattern parsing

  * --------------------------------------------------------- *
  FUNCTION init()
    * Note: Parsing will be done in the order in which the parse patterns are added
    * in some implementations, we may wish  to drive this with meta data
  
    * arguments are addParse( tcPattern, tcReplacement, [tlGlobal] , [tlIgnoreCase], [tlPushToStack] )

    *-- "Heading" parsers. H2, H3, and H4. Note: H4 must be checked first, etc.
    this.addParse( "((^|\n)(\s)*)\[\[\[\[([^\s\*].*?)\]\]\]\]",[$1<h4>$4</h4>],           .t., .t., .f. )   
    this.addParse( "((^|\n)(\s)*)\[\[\[([^\s\*].*?)\]\]\]",    [$1<h3>$4</h3>],           .t., .t., .f. )   
    this.addParse( "((^|\n)(\s)*)\[\[([^\s\*].*?)\]\]",        [$1<h2>$4</h2>],           .t., .t., .f. )   
    * Usage:
    * If any new line starts with (optional white space and then) [[, [[[, or [[[[, followed by any
    * content on the same line that doesn't start with a space or asterisk, followed by an equal number of
    * closing "]" symbols, treat as a metaphor for a level 2, 3, or 4 heading (<h?> tag).
    * [NOTE: We don't do <h1>, because [] is likely intended as something else, and there should be only one <h1>
    * tag ever on a page, which is likely coming from a different source anyway. And we don't do <h5> and <h6>
    * because of frequency-vs-performance trade-off.]

    *-- Emphasis and strikeout inline parsers:
    this.addParse( "(^|\n|\W)-{2,}([^\s|-].*?)-{2,}(\W|$)",    [$1<strike>$2</strike>$3], .t., .t., .f. )
    this.addParse( "(^|\n|\W|_)\*([^\s|\*].*?)\*(_|\W|$)",     [$1<strong>$2</strong>$3], .t., .t., .f. )
    *-- Same as above, but with 2 or more ** 
    this.addParse( "(^|\n|\W)\*{2,}([^\s].*?)\*{2,}(\W|$)",    [$1<strong><em>$2</em></strong>$3], .t., .t., .f. )
    this.addParse( "(^|\n|\W)_([^\s|_].*?)_(\W|$)",            [$1<em>$2</em>$3], .t., .t., .f. )  
    *-- Same as above, but with 2 or more _
    this.addParse( "(^|\n|\W)_{2,}([^\s].*?)_{2,}(\W|$)",      [$1<strong><em>$2</em></strong>$3], .t., .t., .f. )
    * Usage (varies somewhat for minor reasons):
    * Words wrapped with matcing symbols have the symbols replaced with HTML tags as follows:
    * --text--  == <s>       (strikethrough)
    *  *text*   == <strong>
    * **text**  == <strong><em>
    *  _text_   == <em>
    * __text__  == <strong><em>

    *-- Superscript and subscript parsers:
    this.addParse( "\^\{(.*?)\}",                              [<sup>$1</sup>],   .t., .t., .f. )  
    this.addParse( "v\{(.*?)\}",                               [<sub>$1</sub>],   .t., .t., .f. )  
    * Usage:
    * Wrapping text like ^{text} or v{text} is interpreted
    * as superscript and subscript.

    *-- "Horizontal Rule" parsers:
    this.addparse( "(^|\n)----(\s?\r|$)",                      [$1<hr size=1>$2], .t., .t., .f. )
    this.addparse( "(^|\n)====(\s?\r|$)",                      [$1<hr size=4>$2], .t., .t., .f. )
    * Usage:
    * At the beginning of any line, if ---- or ==== is followed by (optional white space and then) nothing but an end-of-line, 
    * then substitute an a thin or thick <hr> tag.
  
    RETURN DODEFAULT() 
    
  ENDFUNC && init

ENDDEFINE 
*-- EOC TextDecorationParser


*** ========================================================== ***
DEFINE CLASS L7FixSlashParser as L7MultiParser
* use this parser to change \ to / for urls (so Mozilla-based browsers, others?)
* will work. 
* NOTE: May want to log hits on this parser so content can be fixed properly
* but, on open content sites, this is a nice parser to have as people
* may not be aware MSIE is cutting them slack with \ chars in URI's

  function init()
    *-- <a followed by a set of chars (non-greedy), followed by href=, followed by some chars, followed by >
    * we clean the 5th submatch.
    this.addParse( "(<a )((.|\n)*?)(href=)((.|\n)*?)(>)", [=$1+$2+$4+strtran($5,"\","/")+$7],.t., .t., .f. ) 
    this.addParse( "(<img)((.|\n)*?)(src=)((.|\n)*?)(>)", [=$1+$2+$4+strtran($5,"\","/")+$7],.t., .t., .f. ) 
    
   *[[ do we need <form here ?, probably a bad habit-forming thing.)
  endfunc
enddefine

*** ========================================================== ***
DEFINE CLASS L7Stack AS L7Lightweight
* some parsers and page elements need a stack to hide their content to

  COUNT      = 0      && returns nElements, see count_access
  nSizeInc   = 5      && size by which the array will be redim'ed (avoides redimensioning on each push()
  nArraySize = 0      && the actual size of the array (some elements may be unused)
  nElements  = 0      && the actual size of the stack (always <= nArraySize)

  DIMENSION aStck(1)  && storage

  HIDDEN nElements, aStck, nArraySize

  * --------------------------------------------------------- *  
  function reset()
    DIMENSION THIS.aStck( THIS.nSizeInc )
    *? OPT: could leave dimension as is (maybe with a max) to avoid lots
    *  of redim's next time.
    
    *-- wipe out all previous (if any) values
    store NULL to this.aStck
    
    THIS.nArraySize = THIS.nSizeInc
    this.nElements = 0
    
  endfunc
  
  * --------------------------------------------------------- *
  FUNCTION INIT()
    this.reset()
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION Count_access
    RETURN THIS.nElements
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION isEmpty()
    RETURN THIS.nElements = 0
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION push( tvValue )

    *-- check if we need to take another chunk of memory
    IF THIS.nElements = THIS.nArraySize
      this.nArraySize =  THIS.nArraySize + THIS.nSizeInc 
      DIMENSION THIS.aStck( THIS.nArraySize )
    
    ENDIF
    THIS.nElements               = THIS.nElements + 1
    THIS.aStck( THIS.nElements ) = m.tvValue
    RETURN THIS.nElements
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION pop()

    IF THIS.nElements > 0
      THIS.nElements = THIS.nElements - 1
      RETURN THIS.aStck( THIS.nElements + 1 )
    ELSE
      RETURN []
    ENDIF
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION getElement( tvElement )

    LOCAL lnElement
    lnElement = IIF( VARTYPE( m.tvElement ) # "N", VAL( m.tvElement ), m.tvElement )
    ASSERT (m.lnElement <= this.nElements) AND m.lnElement > 0 MESSAGE this.NAME + [::getElement() tnElement out of bounds]
 
    IF m.lnElement <= THIS.nElements AND m.lnElement > 0
      RETURN THIS.aStck( m.lnElement )
    ENDIF
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION setElement( tnElement, tvValue )
  
    LOCAL lnElement
  
    lnElement = IIF( VARTYPE( m.tvElement ) # "N", VAL( m.tvElement ), m.tvElement )
    ASSERT ( m.lnElement <= this.nElements) AND m.lnElement > 0 MESSAGE this.NAME + [::setElement() tnElement out of bounds]
  
    THIS.aStck( m.lnElement ) = m.tvValue
  ENDFUNC

ENDDEFINE
*-- EOC Stack


*** ========================================================== ***
DEFINE CLASS L7CORLink AS L7Lightweight

* Defines the COR participant's interface, implements base methods

  oSuccessor    =  NULL    && the next link
  lactive       = .T.      && if .t., this link with pass-through

  lerror        = .f.      && very simplistic, a link can be in two states: error or no error
  cErrorMessage = []
  lhaltonerror  = .f.      && if .t., the chain will terminate if one of the links has an error
  
  protected lactive

  * --------------------------------------------------------- *
  FUNCTION hasSuccessor()
    RETURN VARTYPE( THIS.oSuccessor ) = "O"
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION isActive()
    *-- subclasses may override this for more complex rules
    RETURN THIS.lactive
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION garbagecollect()
    IF THIS.hasSuccessor()
      THIS.oSuccessor.garbagecollect()
    ENDIF
    THIS.oSuccessor = NULL
  ENDFUNC
  
    * ------------------------------------------------------------------- *
  FUNCTION DESTROY
    THIS.GarbageCollect()
    DODEFAULT()
  ENDFUNC  && DESTROY

  * --------------------------------------------------------- *
  FUNCTION RELEASE()
    this.garbageCollect()
    RELEASE this
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION executeImplementation( tvParam )
  * implement in subclass
  * tvParam will be passed by reference
  ENDFUNC
 
  * --------------------------------------------------------- *
  FUNCTION beforeExecute( tvParam )
  ENDFUNC
  
  * --------------------------------------------------------- *
  FUNCTION afterExecute( tvParam ) 
  ENDFUNC
  
  * --------------------------------------------------------- *
  FUNCTION execute( tvParam )
  * The main invoker, template method
  * tvParam must always be passed by reference
       
    this.lError = .f.   && innocent until proven guilty
    
    if this.beforeExecute( @m.tvParam )
      if this.isActive() and not (this.lerror and this.lhaltonerror)
       this.lError = not this.executeImplementation( @m.tvParam )
      endif
    endif   
    this.afterExecute( @m.tvParam )
 
    if not (this.lerror and this.lhaltonerror)  && we may block the rest of the chain after the first error
      IF THIS.hasSuccessor()
        THIS.oSuccessor.execute( @m.tvParam )
      endif
    endif
     
    * a hook for special behaviour at the end of the chain
    * NOTE: This fires even if we are in an error condition
    this.onEndOfChain( @m.tvParam )
       
    return NOT this.lError
 
  ENDFUNC  && execute()

  * --------------------------------------------------------- *
  FUNCTION onEndOfChain( tvParam )
  * fires post execute only if this link is the last in the chain
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION linkCount( tnCount )
  * returns the total number of links in the chain from this point on
    IF VARTYPE( m.tnCount ) #"N"
      tnCount = 1
    ELSE
      tnCount = m.tnCount + 1
    ENDIF
    
    IF THIS.hasSuccessor()
      RETURN THIS.oSuccessor.linkCount( m.tnCount )
    ELSE
      RETURN m.tnCount
    ENDIF
  
  ENDFUNC && linkCount
  
  * --------------------------------------------------------- *
  FUNCTION setAll( tcPropertyName, tvValue)
  *-- set property value for all links in the chain

    THIS.&tcPropertyName = m.tvValue
    IF THIS.hasSuccessor()
      THIS.oSuccessor.SETALL( m.tcPropertyName, m.tvValue )
    ENDIF  
  ENDFUNC
  
  * --------------------------------------------------------- *
  FUNCTION setAllx( tcPropertyName, tvValue, tcExpression )
  *-- set property value for links for which tcExpression evals to true
    LOCAL llcheckExpression, lcExpression
  
    ASSERT (VARTYPE( m.tcExpression ) $ "C") MESSAGE this.NAME + [::setAllx() expects character tcExpression ]
  
    IF EVALUATE( m.tcExpression )
      THIS.&tcPropertyName = m.tvValue
    ENDIF

    IF THIS.hasSuccessor()
      THIS.oSuccessor.setAllx( m.tcPropertyName, m.tvValue, m.tcExpression )
    ENDIF
  ENDFUNC  && setAllx
  
  * --------------------------------------------------------- *
  FUNCTION setSuccessor( tvSuccessor )
  LOCAL loSuccessor
  
    IF VARTYPE( m.tvSuccessor ) = "C"
      loSuccessor = CREATEOBJECT( m.tvSuccessor )
    ELSE
      loSuccessor = m.tvSuccessor
    ENDIF

    ASSERT (VARTYPE( m.loSuccessor ) = "O") MESSAGE this.Name + [:: setSuccessor() requires an object or a class name ]
  
    IF THIS.hasSuccessor()
      RETURN THIS.oSuccessor.setSuccessor( m.loSuccessor )
    ELSE
      *-- return a reference as we might want to tweak this link after adding it
      THIS.oSuccessor = m.loSuccessor
      RETURN m.loSuccessor
    ENDIF

  ENDFUNC  && setSuccessor

ENDDEFINE
*--EOC COR Link

*** ========================================================== ***
DEFINE CLASS L7CRParser as L7BaseParser
* provides <br's> for carriage returns
* due to lack of good lookbehind assertion grouping in vbscript, we'll 
* take the ALINES() route here.

  * tags we don't want to put <br> behind
  * use lowercase, both opening and closing tags will be blocked
  * ie, <table and </table will be blocked if |table| appears in the list
  cDisallowedEndingTags = [|table|tr|td|th|ol|li|ul|hr|br|ig|p|div|!|h1|h2|h3|h4|-|script|pre|col|colgroup|thead|tbody|]
  
  *-- we won't add any <br> in blank lines following these tags, note you must
  *-- include both open and close tags here as needed
  cDisallowedFollowTags = [table,tr,/tr,td,/td,th,/th,p,/p]
  *-- note that we do allow <br>s after tables
  
  cBRTag = L7BR 
  
  * --------------------------------------------------------- *  
  FUNCTION isPatternPresent( tcText )
  RETURN !EMPTY( m.tcText )
  
  * --------------------------------------------------------- *
  FUNCTION tagOKForBR( tcTag )
    LOCAL llRetVal
    llRetval = ! ( [|] + ALLTRIM( LOWER( m.tcTag ) ) + [|] $ this.cDisallowedEndingTags )
    RETURN m.llRetVal
  ENDFUNC && tagOKForBR
  
  * --------------------------------------------------------- *  
  FUNCTION parse( tcText )
    *-- note: tcText is passed by reference

    LOCAL lnTotLines, laLines(1), lcLine, lcParsedText, llValidLine, lcTag, lnLastTag, lnLine, lcWholeTag, lcBRTag
    
    lcBRTag         = this.cBRTag
    lnTotLines      = ALINES( m.laLines, m.tcText )
    lcParsedText    = []
    lcTag           = [~]    && placeholder for first pass.
    lcWholeTag      = [~]
    
    FOR m.lnLine= 1 TO m.lnTotLines
    
      lcLine      = laLines( m.lnLine )  
      llValidLine = .t.  
      llValidLine = NOT( EMPTY( m.lcLine ) AND ;
                    m.lcWholeTag $ this.cDisallowedFollowTags )
      *---since the HTMLIgnore parser may be yanking things we should not
      *   be placing a <br> behind, we need to avoid the placeholder (left by HTMLIgnore)
      *   this can be a problem if *other* parsers yank stuff we *do* want to <br>
      *   Testing changes.. HTMLIgnore now leaves the tag tailings so CR parser can decide.
*      llValidLine = llValidLine and NOT( RIGHT( ALLTRIM( m.lcLine ), 1) = PLACEHOLDER_DELIM )
      
      IF llValidLine AND RIGHT( ALLTRIM( m.lcLine ), 1) = [>] 
          *-- does line end with a tag?    
          lnLasttag = OCCURS([<], m.lcLine)
        
          IF m.lnLastTag>0
            lcTag = STREXTRACT( ALLTRIM(m.lcLine) + [*], [<], [>*], m.lnLastTag)   
            lcWholeTag = GETWORDNUM( m.lcTag, 1) && includes possible leading / char 
            lcTag = IIF( m.lcTag = "/", SUBSTR( m.lcTag, 2 ), m.lcTag)
            
            lcTag = IIF( m.lcTag = "!", LEFT( m.lcTag, 1 ), m.lcTag)
            lcTag = GETWORDNUM( m.lcTag, 1)
            llValidLine = this.tagOKforBR( m.lctag )
          ENDIF
      ENDIF 
        
      *-- llValidLine will be .t. if this doesn't end with a bad tag
      lcParsedText = m.lcParsedText + ;
                     m.lcLine + ;
                     IIF( m.llValidLine and m.lnLine < m.lntotLines, m.lcBRTag, [] ) + CRLF 
          
    ENDFOR  
    tcText = m.lcParsedText

  ENDFUNC && parse
    
ENDDEFINE 
*-- EOC CRParser



*** ========================================================== ***
DEFINE CLASS L7DynamicMarkers as L7MultiParser
* This chain is set up for "on save" parsing. IOW, it does things
* that would be performed on raw submissions prior to save
*

  * --------------------------------------------------------- *
  FUNCTION getDate()
    LOCAL ldRet

    ldRet = DATE()
    
    *-- fix the date so tests don't constantly fail
    #IF L7_PARSER_DEBUG
      ldRet = DATE(2002,8,01)
    #ENDIF
    
    RETURN m.ldRet
  ENDFUNC
  
  * --------------------------------------------------------- *
  FUNCTION getDateDiff( tcJulianDay )
    LOCAL ldCurrentDate, lnJD
    
    ASSERT VARTYPE( m.tcJulianDay ) = "C" MESSAGE this.name + [::getDateDiff(c) expects character parameter ]
    
    IF VARTYPE( m.tcJulianDay ) # "C"
      tcJulianDay = "0"
    ENDIF
    
    lnJD   = VAL( m.tcJulianDay )
    
    ldCurrentDate = this.getDate()

    RETURN VAL( SYS(11, m.ldCurrentDate ) ) - m.lnJD
    
  ENDFUNC
  
  * --------------------------------------------------------- *
  FUNCTION init()
    * NOTE: this is a "multi-context" parser, these sub-parses will fire (or not) depending on the context..
    
    this.addParse( "(^|\n|\W)(#d#)(\W|$)",         "=$1+'['+this.getDateMarker()+']'+$3", .t., .t., .f.,[SAVE])
    
    this.addParse( "(^|\n|\W)(#n#)(\W|$)",           [=$1+"<new="+SYS(11,this.getDate())+">"+$3], .t., .t., .f., [SAVE])
    this.addParse( "(^|\n|\W)(\<new=(\d+)\>)(\W|$)", [=$1+this.getNewMarker($3)+$4], .t., .t., .f., [VIEW])
    this.addParse( "(^|\n|\W)(#u#)(\W|$)",               [=$1+"<updated="+SYS(11,this.getDate())+">"+$3], .t., .t., .f., [SAVE])
    this.addParse( "(^|\n|\W)(\<updated=(\d+)\>)(\W|$)", [=$1+this.getUpdatedMarker($3)+$4], .t., .t., .f., [VIEW])
    
    RETURN DODEFAULT()
  
  ENDFUNC
 
  * --------------------------------------------------------- *
  FUNCTION getDateMarker()
  * returns <span class="updatedX">[Updated]</span>, or nothing.
  *? CSS ALERT
    * this is a hook incase we want to do special date formatting
    RETURN TRANSFORM( this.getDate() )
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION getNewMarker( tcJulianDay )
  * returns <span class="newX">[NEW]</span>, or nothing.
  *? CSS ALERT
    LOCAL lcStr, lnDiff
  
    lnDiff = this.getDateDiff( tcJulianDay )

    IF BETWEEN( m.lnDiff, 0, 8 )
      lcStr = [<span class="new] + ALLTRIM( STR( m.lnDiff ) ) + '">[NEW]' + [</span>]
    ELSE 
      lcStr = []
    ENDIF
    RETURN m.lcStr
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION getUpdatedMarker( tcJulianDay )
  * returns <span class="UpdatedX">[Updated]</span>, or nothing.
  *? CSS ALERT
    LOCAL lcStr, lnDiff
  
    lnDiff = this.getDateDiff( m.tcJulianDay )

    IF BETWEEN( m.lnDiff, 0, 8 )
      lcStr = [<span class="updated] + ALLTRIM( STR( m.lnDiff ) ) + '">[UPDATED]' + [</span>]
    ELSE 
      lcStr = []
    ENDIF
    RETURN m.lcStr
  ENDFUNC

ENDDEFINE


*==============================================================================
*==============================================================================
* PARSING SUPPORT CLASSES
*==============================================================================
*==============================================================================

*** ========================================================== ***
DEFINE CLASS L7BaseParser AS L7CORLink

  cCurrentContext    = []               && set by setContext() to determine which parsers fire and how they fire
  cValidContextSet   = []               && the context(s) for which this parser is active
  cDefaultContext    = []               && the context under which this parser typically operates
  
  oRegExp            = NULL             && regular expression provider, may be shared over a chain
  oStack             = NULL             && stack to store extracted patterns (if needed) likely to be shared
                                        && over an entire chain and emptied by last link
                                        
  cRegExpClass = L7_REGEXP_CLASS        && the class to use by default for oRegExp
  cStackClass  = "L7Stack"              && the class to use by default for oStack

  cPattern           = .f.             && the pattern to look for (set to .f. to identify parsers that don't use patterns at all)
  cReplacement       = []               && the replacement (may include $1, $2 for submatches, and, may include a & prefix for macros)

  lGlobal            = .T.              && find all occurances by default
  lIgnoreCase        = .T.              && obvious
  lMultiLine         = .F.              
  lPushToStack       = .F.              && if .t., matches will be placed on the stack and placeholders left behind in text
  cPlaceHolderDelim  = L7_PARSER_PLACEHOLDER_DELIM && used to pad stack placeholders
  
  lisAnchored        = .f.              && will be .t. if link is anchored, ie, is owned by an L7ParseChainAnchor object
  
  *-- status logging properties       
  lLog               = .f.
  nHits              = 0
  nCumulativeHits    = 0
  nSeconds           = 0
  nCumulativeSeconds = 0
  nStartSeconds      = 0
  cStartText         = []
  cEndText           = []
  lVerbose           = .f.              && if .t., before and after text will also be displayed in status, caution: can eat memory fast this way.
 * cLogPlaceholder    = []               && used to find the log in the stack after parsing

  PROTECTED cCurrentContext, oRegExp, oStack, lisAnchored
  
  * --------------------------------------------------------- *
  FUNCTION garbagecollect()
 
    IF THIS.hasSuccessor()
      THIS.oSuccessor.garbagecollect()
    ENDIF
    THIS.oSuccessor = NULL
    this.oRegExp    = NULL
    this.oStack     = NULL
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION copyValueFromConfig( toConfig, tcConfigProp, tcThisProp )
    *-- Imputes config object property value to indicated this.prop.
    *   this function would be called from ::Config()
     
    *-- allow local property name to be assumed
    IF VARTYPE( tcThisProp ) # "C"
      tcThisProp = tcConfigProp
    ENDIF
    
    IF PEMSTATUS( toConfig,tcConfigProp, 5)
      this.&tcThisProp = toConfig.&tcConfigProp
    ENDIF
  
  ENDFUNC

  * --------------------------------------------------------- *  
  FUNCTION configImplementation(toConfig)
    *-- to be overridden in application subclasses...
    * the idea is to use this.imputeConfigValue() copy properties from toConfig into the parser
    *? consider an "auto config" that traverses all properties and looks for matches (exclude NAME etc)
  ENDFUNC
  
  * --------------------------------------------------------- *  
  FUNCTION config( toConfig )
    WITH this
      .configImplementation( toConfig )
      IF .hasSuccessor()
        this.oSuccessor.config(  toConfig )
      ENDIF
    ENDWITH
    
  ENDFUNC  
    
  * --------------------------------------------------------- *
  FUNCTION setContext( tcContext )
    THIS.SETALL( "cCurrentContext", m.tcContext )
  ENDFUNC
  
  * --------------------------------------------------------- *
  FUNCTION resetContext()
    this.cCurrentContext = this.cDefaultContext
    IF this.hasSuccessor()
      this.oSuccessor.resetContext()
    ENDIF
  ENDFUNC
  
  * --------------------------------------------------------- *
  FUNCTION isValidContext()
   
    *-- no valid contexts, we'll do anything
    *[[ consder using NULL for this.
    IF empty( this.cValidContextSet )
      RETURN .t.
    ENDIF
    
    IF !EMPTY( this.cCurrentContext ) && we've been given a specific context
      RETURN ([,] + ALLTRIM( this.cCurrentContext ) + [,]) $ ([,] + ALLTRIM( this.cValidContextSet ) + [,])
    ELSE
      RETURN ([,] + ALLTRIM( this.cDefaultContext ) + [,]) $ ([,] + ALLTRIM( this.cValidContextSet ) + [,])
    ENDIF
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION onEndOfChain( tcText )
    IF this.oStack.count>0
      THIS.clearStack( @m.tcText )
    ENDIF
  ENDFUNC
  
  * --------------------------------------------------------- *
  FUNCTION buildPlaceholder( tcToken )
    *-- wraps stack element placeholders
    RETURN THIS.cPlaceHolderDelim + m.tcToken + THIS.cPlaceHolderDelim
  ENDFUNC
  
  * --------------------------------------------------------- *
  FUNCTION pushToStack( tcReplacement )
    LOCAL lcRetVal, lnElements
    lnElements = THIS.oStack.PUSH( m.tcReplacement )
    lcRetVal   = THIS.buildPlaceholder( ALLTRIM( STR( m.lnElements ) ) )
    RETURN m.lcRetVal
  ENDFUNC



  * --------------------------------------------------------- *
  FUNCTION clearStack( tcText )
  * tcText passed by reference
  *-- spew stack contents back into placeholder positions
  *   normally only called on the last parser in the chain

    LOCAL lcText, llPushToStack, lcPlaceHolderPattern, lnIts

    lcPlaceHolderPattern = THIS.buildPlaceholder([(\d+)])

    *? kluge, we don't want replaceex() below to push to stack
    llPushToStack     = THIS.lPushToStack
  
    THIS.lPushToStack = .F.

    lnIts = 0
  
    *-- set the maximum nesting level to 10
    *   because stack contents may include placeholders (due to nested extractions)
    *   we may need to take multiple passes to clear all the placeholders
    *? OPT: because what is being done here is relatively simple, we could
    *  use native VFP (rather than regexp in replaceex() ) to get the job done
    DO WHILE THIS.m( m.tcText, m.lcPlaceHolderPattern ) AND m.lnIts < 10
      THIS.replaceex( @m.tcText, m.lcPlaceHolderPattern, [=this.oStack.getElement( $1 )] )
      lnIts = lnIts + 1
    ENDDO
    
    *-- reset the stack to clear memory
    
    this.oStack.reset()
    
    *-- reset to stored value
    THIS.lPushToStack = llPushToStack
    
  ENDFUNC  && clearStack
  
  * --------------------------------------------------------- *
  FUNCTION getLog()
    LOCAL lcStr
    
    lcStr = PADR( this.name, 20 ) + [|] + [H: ]   + PADR( this.nHits, 4 )           + ;
                              [|CH: ] + PADR( this.nCumulativeHits, 6 ) + ;
                              [|CX: ]  + PADR( this.cCurrentContext, 10 )        + ;
                              [|DX: ] + PADR( this.cDefaultContext, 10 ) + ;
                              [|VCX: ] + PADR( this.cValidContextSet, 20 )
                              
    IF this.lverbose AND this.nHits > 0
      LOCAL lcTextDelim
      lcTextDelim = + REPLICATE( "#",20) 
      lcStr = lcStr + CRLF + m.lcTextDelim + encodeForHTML(this.cStartText) + CRLF + ;
                             m.lcTextDelim + CRLF + ;
                             encodeforHTML(this.cEndText) + m.lcTextDelim + CRLF
                             
      *? consider getting even more verbose here with property settings, stack contents etc...
    ENDIF
    
    RETURN m.lcStr 
    
  ENDFUNC
  
  * --------------------------------------------------------- *
  FUNCTION detectLoggingRequest( tcText )
  * checks the text for logging flags and sets chain members llLog and lVerbose accordingly
  
    *-- note: a bit of an extra cost here for long unachored chains
    IF !this.lisAnchored AND ATC("showparselog", m.tcText ) > 0
  
      *-- first, turn on logging as needed
      DO CASE 
        CASE ATC("[showparselog]", m.tcText ) + ATC("[showparselog verbose]", m.tcText ) > 0
          this.setAll( [lLog] , .t. ) && do all of them
        CASE ATC("[showparselog ", m.tcText ) > 0
          PRIVATE pcTempText
          LOCAL lnPos
          lnPos      = ATC("[showparselog ", m.tcText )
          pcTempText = SUBSTR(m.tcText,m.lnPos)  && not, perfect but probably good enough
          this.setAllx( [llog], .t. , [' '+LOWER(this.NAME) $ LOWER( m.pcTempText ) ])
        otherwise
          error [L7Parser::detectLoggingRequest() unexpected status detected in set logging.]
      ENDCASE
      
      *-- second, turn on verbose as needed
      DO CASE
        CASE  ATC("[showparselog verbose", m.tcText) > 0
         this.setAll( [lVerbose], .t. )
        CASE ATC("verbose", m.tcText) > 0         
           PRIVATE pcTempText 
           pcTempText = m.tcText
           *-- conditionally set verbose on particular classes that have been requested
           this.setAllx( [lVerbose], .t., "LOWER(this.NAME) +' verbose' $ LOWER( m.pcTempText )" )
        otherwise
	       * no extra work to do
        ENDCASE
     ENDIF
  ENDFUNC
  
  * --------------------------------------------------------- *
  FUNCTION beforeExecute( tcText )
    this.detectLoggingRequest( m.tcText ) 
  ENDFUNC
   
  * --------------------------------------------------------- *
  FUNCTION getObjectRef( tcDefaultClass, tvObjectOrClass, tcModule )
    LOCAL loRet
    *? possibly hook out to optional factory object here

    DO CASE
      CASE VARTYPE( m.tvObjectOrClass ) = "O"
        loRet = m.tvObjectOrClass
     CASE VARTYPE( m.tvObject ) = "C"
       IF VARTYPE( m.tcModule ) = "C"
         loRet = NEWOBJECT( m.tvObjectOrClass, m.tcModule )
       ELSE
         loRet = CREATEOBJECT( m.tvObjectOrClass )
       ENDIF
     OTHERWISE
       loRet = CREATEOBJECT( m.tcDefaultClass )
     ENDCASE
    RETURN m.loRet
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION setStack( tvStack, tcModule )
    THIS.oStack = THIS.getObjectRef( THIS.cStackClass , m.tvStack, m.tcModule )
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION setRegExp( tvRegExp, tcModule )
    THIS.oRegExp = THIS.getObjectRef( THIS.cRegExpClass, m.tvRegExp, m.tcModule )
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION configRegExp() as boolean
    WITH THIS.oRegExp
      .GLOBAL     = THIS.lGlobal
      .ignorecase = THIS.lIgnoreCase
      .multiline  = THIS.lMultiLine
    ENDWITH
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION m( tcText, tcPattern ) as boolean
    * same as perl's m
    this.configRegExp()
    THIS.oRegExp.PATTERN = m.tcPattern
    RETURN THIS.oRegExp.test( m.tcText )
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION replaceex( tcText, tcSearchPattern, tcReplacePattern ) as integer
  * mimics perl's 's' function
  * take note: nested calls to replaceex() will fail if 
  * RegExp::Execute() fails to return NEW match collections, but as of 5.5 it does, and seems
  * unlikely it would stop doing so
  * also, tcText must be passed by reference

    LOCAL loRE, lcText, lnShift, lcReplacePattern, lcCommand
    LOCAL loMatch, loMatches, lnI, lcSubMatch, llevaluate, lcMatchDelim, lnHits
    LOCAL lcReplacement

    lnHits           = 0
          
    IF EMPTY( m.tcText )
      RETURN 0
    ENDIF
  
    this.configRegExp()
  
    loRE         = THIS.oRegExp
    loRE.PATTERN = m.tcSearchPattern
  
    *--- are we evaluating?
    IF tcReplacePattern = [=]
      llEvaluate        = .t.
      lcReplacePattern  = SUBSTR( m.tcReplacePattern, 2 )
    ELSE
      lcReplacePattern  = m.tcReplacePattern
    ENDIF
  
    *-- check for external calls
    IF this.llog OR llevaluate OR this.lPushToStack
      * in order to get counts to log, we always have to do this...
    
      lcText           = m.tcText
      lnShift          = 1
*      lcMatchDelim     = IIF( m.llEvaluate, [],[])  && depreciate this..
      loMatches        = loRE.execute( m.lcText )


      *? take note: fox bug, can't close datasession after this?
      * potential fix, use for lnK=1 to count instead...
      
      FOR EACH loMatch IN m.loMatches
      
        lnHits    = m.lnHits + 1  
        lcCommand = m.lcReplacePattern
        *? consider start with 0, the 0th submatch is the entire match
        lnI       = 1  
            
        FOR EACH lcSubMatch IN loMatch.SubMatches
       
          *-- we can't have cr's in code we are going to evaluate
          *   the typical situation is [&"<pre>"+encodeforHTML("$1$2$4")+"</pre>"]
          *   the linefeeds need to be escaped prior to the function call
          IF llEvaluate 
            * "escape" the string we are about to use in an evaluation.
            * it is important to escape due to possible delim chars (like ", ' etc)
            * malicious content, or VFP line-length violations.
            lcSubMatch = [this.oStack.getElement(]+ALLTRIM(STR(this.oStack.push( m.lcSubMatch )))+[)]
          ENDIF
          
          lcCommand = STRTRAN( m.lcCommand, "$" + ALLTRIM( STR( m.lnI ) ) , m.lcSubMatch)
          lnI = m.lnI + 1
        ENDFOR
     
        IF m.llEvaluate
          *[[ consider try/catch here
          lcReplacement = EVALUATE( m.lcCommand )
        ELSE
          lcReplacement = m.lcCommand
        ENDIF 

        IF THIS.lPushToStack
          *-- push the replacement to the stack and get a placeholder to use
          lcReplacement = THIS.pushToStack( m.lcReplacement )
        ENDIF

        lcText  = STUFF( m.lcText, loMatch.FirstIndex + m.lnShift, m.loMatch.LENGTH, m.lcReplacement )
        lnShift = m.lnShift + LEN( m.lcReplacement ) - m.loMatch.LENGTH

      ENDFOR
    
    ELSE
      *--- nothing fancy, just a raw replace
      * note: this code is not active if we are logging
      * it is faster, but we can't count hits...
      lcText = THIS.oRegExp.REPLACE( m.tcText, m.tcReplacePattern )
      
    ENDIF
    tcText = m.lcText && since we are passing tcText by reference
    RETURN m.lnHits
 
  ENDFUNC
  
  * --------------------------------------------------------- *
  FUNCTION beforeParse( tcText ) AS boolean
  *-- hook. tcText passed by ref

    this.nHits         = 0 
    
    IF this.lLog 
      this.nStartSeconds = SECONDS()  
      this.cStartText    = m.tcText
    ENDIF
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION afterParse( tcText ) AS boolean
  *-- hook. tcText passed by ref
    *?! this may not be good default functionality...
    alltrimX( @m.tcText )
    
    IF this.lLog
      this.nSeconds           = this.nSeconds + SECONDS() - this.nStartSeconds
      *-- this.nHits is incremented in this.parse() 
      this.cEndText           = m.tcText
      this.nCumulativeHits    = this.nCumulativeHits    + this.nHits
      this.nCumulativeSeconds = this.nCumulativeSeconds + this.nSeconds
      tcText = m.tcText + this.pushtostack([<pre style="font-size: xx-small">]+this.getLog()+[</pre>] )
    ENDIF
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION parse( tcText ) AS boolean
      *-- hook, the meat, tcText passed by ref. Likely to be overwritten in subclass
      this.nHits = this.nHits + THIS.replaceex( @m.tcText , THIS.cPattern, THIS.cReplacement )
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION isActive() AS boolean
    RETURN THIS.isValidContext() AND DODEFAULT()
  ENDFUNC
  
  * --------------------------------------------------------- *
  FUNCTION isPatternPresent( tcText ) as Boolean
   
    if vartype( this.cPattern ) = "L"
      *-- we'll not even check in this case, thus allowing subclasses
      *   that do not use the pattern property to fire the parse() method by default
      return .t.
    endif
    
    IF EMPTY( this.cPattern ) OR EMPTY( m.tcText )
      RETURN .f.
    ENDIF
    
    RETURN this.m( m.tcText, this.cPattern )
      
  ENDFUNC

  * --------------------------------------------------------- *
  PROTECTED FUNCTION executeImplementation( tcText ) as boolean
  * main template method for the parsers
  * implementation code goes into the hooks called here
     
    *-- make sure we've got what we need.
    if isnull(this.oRegExp) 
      this.setRegExp()
    endif
    
    if isnull(this.oStack)
      this.setStack()
    endif
    
    *-- hook.
    THIS.beforeParse( @m.tcText )
    
    *-- check we've got work to do
    IF this.isPatternPresent( m.tcText )  && allow for quick bail 
      *-- hook.
      THIS.parse( @m.tcText )
    ENDIF
    
    *-- hook.
    THIS.afterParse( @m.tcText )
  
    return not this.lError

  ENDFUNC
  
ENDDEFINE
*-- EOC Base Parser
  
*** ========================================================== ***
DEFINE CLASS L7MultiParser as L7BaseParser
* extends the L7Baseparser with addParse() to add multiple simple parsing patterns
* use this for groups of simple parsing tasks that are related

  nParseCount = 0

  DIMENSION aParses(1)
  HIDDEN aParses

  * --------------------------------------------------------- *
  FUNCTION addParse( tcPattern, tcReplacement, tlGlobal, tlIgnoreCase, tlPushToStack, tcContextSet )
    LOCAL lnPcount, llGlobal, llIgnoreCase, llPushToStack, lnPFlagcount,lcContextSet

    lnPFlagcount = PCOUNT() - 2 && number of flags sent
    
    llGlobal      = IIF( m.lnPFlagcount>=1, m.tlGlobal,      this.lGlobal      )
    llIgnoreCase  = IIF( m.lnPFlagcount>=2, m.tlIgnoreCase,  this.lIgnoreCase  )
    llPushToStack = IIF( m.lnPFlagcount>=3, m.tlPushToStack, this.lPushToStack )
    lcContextSet  = IIF( m.lnPFlagcount>=4, m.tcContextSet,  this.cValidContextSet )
    
    this.nParseCount = this.nParseCount + 1 
    DIMENSION this.aParses( this.nParseCount, 6 )
  
    *-- take note: executeImplementation() depends on column order
    this.aParses( this.nParseCount , 1 ) = m.tcPattern
    this.aParses( this.nParseCount , 2 ) = m.tcReplacement 
    this.aParses( this.nParseCount , 3 ) = m.llGlobal
    this.aParses( this.nParseCount , 4 ) = m.llIgnoreCase
    this.aParses( this.nParseCount , 5 ) = m.llPushToStack
    this.aParses( this.nParseCount , 6 ) = m.lcContextSet
    
  ENDFUNC  && addParse
  
  * --------------------------------------------------------- *
  FUNCTION resetParses()
    this.nParseCount = 0
    DIMENSION this.aParses(1)
    this.aParses(1) = .f.
  ENDFUNC
  
  * --------------------------------------------------------- *
  FUNCTION executeImplementation( tcText )
    LOCAL lnK,                 ;
          lcOldPattern,        ;
          lcOldReplacement,    ;
          llOldGlobal,         ;
          llOldPushToStack,    ;
          llOldIgnoreCase,     ;
          lcOldValidContextSet
  
    ASSERT VARTYPE( m.tcText ) = "C" MESSAGE this.Name + [::executeImplementation() expects a character argument ]
  
      *-- make sure we've got what we need.
    if isnull(this.oRegExp) 
      this.setRegExp()
    endif
    
    if isnull(this.oStack)
      this.setStack()
    ENDIF
    
    IF this.nParseCount > 0 
  
      lcOldPattern         = this.cPattern
      lcOldReplacement     = this.cReplacement
      llOldGlobal          = this.lGlobal
      llOldPushToStack     = this.lPushToStack
      llOldIgnoreCase      = this.lIgnoreCase
      lcOldValidContextSet = this.cValidContextSet
      
      this.beforeParse( @m.tcText )
     
      FOR lnK = 1 TO this.nParseCount
    
        *-- take note: the column order is established in addParse()
        this.cPattern         = this.aParses( m.lnK, 1 )
        this.cReplacement     = this.aParses( m.lnk, 2 )
        this.lGlobal          = this.aParses( m.lnK, 3 )
        this.lIgnoreCase      = this.aParses( m.lnK, 4 )
        this.lPushToStack     = this.aParses( m.lnK, 5 )
        this.cValidContextSet = this.aParses( m.lnK, 6 )

        IF this.isValidContext()
          this.parse( @m.tcText )
        ENDIF    
      ENDFOR
    
      this.afterParse( @m.tcText )
    
      *-- replace divots
      this.cPattern         = m.lcOldPattern
      this.cReplacement     = m.lcOldReplacement
      this.lPushToStack     = m.llOldPushToStack
      this.lGlobal          = m.llOldGlobal
      this.lIgnoreCase      = m.llOldIgnoreCase
      this.cValidContextSet = m.lcOldValidContextSet
      
    ELSE
      *-- we have a simple this.cPattern, this.cReplacement based parsing job to do
      *   SimpleParser can handle this.
      DODEFAULT( @m.tcText )
    ENDIF
  
    ASSERT VARTYPE( m.tcText ) = "C" MESSAGE this.Name + [::executeImplementation() changing arguement type]
  
    RETURN
  ENDFUNC  && MultiParser::executeImplementation
ENDDEFINE
*-- EOC MultiParser 

*** ========================================================== ***
DEFINE CLASS L7ParseChainManager as L7Lightweight
*  Manages multiple chains, keeping them cached for repeated use
*  this class is meant to be hung of a persistent object (like oApp) 

nChainCount = 0
DIMENSION aChains(1)

PROTECTED nChainCount, aChains

  * --------------------------------------------------------- *
  FUNCTION getObjectRef( tcDefaultClass, tvObjectOrClass, tcModule )
    LOCAL loRet
    *? possibly hook out to optional factory object here

    DO CASE
      CASE VARTYPE( m.tvObjectOrClass ) = "O"
        loRet = m.tvObjectOrClass
     CASE VARTYPE( m.tvObject ) = "C"
       IF VARTYPE( m.tcModule ) = "C"
         loRet = NEWOBJECT( m.tvObjectOrClass, m.tcModule )
       ELSE
         loRet = CREATEOBJECT( m.tvObjectOrClass )
       ENDIF
     OTHERWISE
       loRet = CREATEOBJECT( m.tcDefaultClass )
     ENDCASE
    RETURN m.loRet
  ENDFUNC

  * --------------------------------------------------------- *  
  FUNCTION config( toConfig )
    IF VARTYPE( toConfig ) = "O"
      FOR EACH loChain IN this.aChains
        loChain.config( toConfig )
      ENDFOR
    ENDIF
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION addChain( tvChain, tcName )
    LOCAL loChain
    
    ASSERT VARTYPE( tvChain ) $ "CO" MESSAGE this.name + [::getChainByName( c ) expects character or object for 1st param ]
    
    IF NOT VARTYPE( tvChain ) $ "CO"
      RETURN .f.
    ENDIF
    
    IF VARTYPE(tvChain) = "C"
      *? factory alert
      loChain = this.getObjectRef( tvChain )
    ELSE
      loChain = m.tvChain
    ENDIF
 
    *-- give this chain it's name (will default to class name if not passed)
    *? perhaps check for uniqueness here?
    IF VARTYPE( tcName ) = "C"    
      loChain.Name = m.tcName
    ENDIF
  
    *-- forge the chain now, to have it hot and ready for the first hit
    *  might just be a basic chain, not anchored, so we check for the forgeChain method
    *? kluge alert... think about this some more...
    IF PEMSTATUS( loChain,[forgeChain],5)
      loChain.forgeChain()
    ENDIF
          
    this.nChainCount = this.nChainCount + 1 
    DIMENSION this.aChains( this.nChainCount )
    this.aChains( this.nChainCount ) = m.loChain
    
    RETURN m.loChain
    
  ENDFUNC
  

  * --------------------------------------------------------- *
  FUNCTION getChainByName ( tcName )
    LOCAL loChain
    
    ASSERT VARTYPE( m.tcName ) = "C" MESSAGE this.name + [::getChainByName( c ) expects character for 1st param ]
        
    FOR EACH loChain IN this.aChains
      IF ALLTRIM(UPPER(loChain.Name)) == ALLTRIM(UPPER( m.tcName))
        RETURN m.loChain
      ENDIF
    ENDFOR
    RETURN null

  ENDFUNC
    
    
  * --------------------------------------------------------- *
  FUNCTION parseByName( tcNames, tcText, tcContext )
    * parses tcText with the chains in TcNames, with context tcContext
    LOCAL laNames(1), lnK, lnTotalNames, llsetSpecialContext, loChain

    ASSERT VARTYPE( m.tcNames ) = "C" MESSAGE this.name + [::parseByName( c, c) expects character for 1st param ]
    ASSERT VARTYPE( m.tcText )  = "C" MESSAGE this.name + [::parseByName( c, c) expects character for 2nd param ]
    
    this.beforeParseByName( m.tcNames, @m.tcText, m.tcContext )
    
    llsetSpecialContext = VARTYPE( m.tcContext ) = "C"
    
    lnTotalNames = ALINES( m.laNames, m.tcNames, .t., [,] )
    
    FOR lnK = 1 TO m.lnTotalNames
      
      loChain = this.getChainByName( laNames( lnk ) )
      
      IF VARTYPE( loChain ) = "O"
        
        IF llsetSpecialContext
          loChain.setContext( m.tcContext )
        ENDIF
        
        loChain.execute( @m.tcText )
      
        IF llsetSpecialContext
          loChain.resetContext() && put each link's context back to default
        ENDIF
              
      ENDIF
      
    ENDFOR  
    
    this.afterParseByName( tcNames, @tcText, tcContext )
    
    *-- take note, we operate on tcText AND return it, that way this function can be called with tcText by 
    *   ref or by value
    RETURN m.tcText 
   
  endfunc
  
  function beforeParseByName( tcNames, tcText, tcContext )
  endfunc
  
  function afterParseByName( tcNames, tcText, tcContext )
  endfunc
  
ENDDEFINE
  
*** ========================================================== ***
DEFINE CLASS L7ParseChainAnchor as L7BaseParser
* a special link that brings one or more links with it
* orchestrates the chain to work coherently

  * --------------------------------------------------------- *
  FUNCTION cPlaceHolderDelim_ASSIGN( tvValue ) as boolean
    *-- ensures all parsers use the same placeholder
    this.setAll( "cPlaceHolderDelim", m.tvValue )
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION forgeChainImplementation()
    *-- build the chain here, possibly from meta data
    RETURN
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION forgeChain()
  *-- builds the chain, implement in subclass, often from meta data

    *-- build the chain
    this.forgeChainImplementation()  
    
    this.setAll("lisAnchored", .t. ) && indicate that all links are anchored
    this.lisAnchored = .f.           && the anchor isn't anchored, doesn't make sense...
    
    *-- instantiate the shared elements
    this.setStack()
    this.setRegexp()
  
    *-- pointer all links to the same oRegExp and oStack
    *? could do this conditionally with setAllx for oRegExp to allow some links to own their own regExp object..
    this.setall("oRegExp", this.oRegExp )
    this.setall("oStack" , this.oStack  )
 
    *-- same placeholder must be used in order to share the stack
    this.setall("cPlaceHolderDelim", this.cPlaceHolderDelim)
         
  ENDFUNC

  * --------------------------------------------------------- *
  FUNCTION execute( tcText )
    *-- check that we have a chain attached
    *-- this allows delayed instantiation of the chain
    IF !this.hasSuccessor()
      this.forgeChain()
    ENDIF
    RETURN DODEFAULT( @m.tcText )
  ENDFUNC
ENDDEFINE
*-- EOC ParseChainAnchor

*** ========================================================== ***
DEFINE CLASS L7Lightweight AS RELATION

  HIDDEN CHILDALIAS
  HIDDEN CHILDORDER
  HIDDEN ONETOMANY
  HIDDEN PARENTALIAS
  HIDDEN RELATIONALEXPR
  
ENDDEFINE